{**********************************************************************
 Package pl_UIB
 This file is part of CodeTyphon Studio (https://www.pilotlogic.com/)
***********************************************************************}

unit alluibregister;

{$I uib.inc}

interface

uses
  Classes,dialogs, Math, Forms, Controls,
  PropEdits, ComponentEditors,
  uib,
  uibdataSet,
  uibdatabaseedit,
  uibtransactionedit,
  uibsqledit;

type
  TUIBDatabaseEditor = class(TComponentEditor)
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

  TUIBTransactionEditor = class(TComponentEditor)
  public
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

  TUIBSQLEditor = class(TComponentEditor)
  public
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

  TUIBStatementSQLProperty = class(TClassProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;

  TUIBDatasetSQLProperty = class(TClassProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;

procedure Register;

implementation

{$R alluibregister.res} //=== ct9999 ==============

procedure Register;
begin

  RegisterComponents('UIB', [TUIBDatabase,
                             TUIBTransaction,
                             TUIBQuery,
                             TUIBBackup,
                             TUIBRestore,
                             TUIBScript,
                             TUIBEvents,
                             TUIBRepair,
                             TUIBSecurity,
                             TUIBConfig,
                             TUIBServerInfo,
                             TUIBDataSet]);

  RegisterComponentEditor(TUIBDatabase, TUIBDatabaseEditor);
  RegisterComponentEditor(TUIBTransaction, TUIBTransactionEditor);

  RegisterComponentEditor(TUIBStatement, TUIBSQLEditor);

  RegisterPropertyEditor(TypeInfo(TStrings), TUIBStatement,'SQL', TUIBStatementSQLProperty);
  RegisterPropertyEditor(TypeInfo(TStrings), TUIBDataset,  'SQL', TUIBDatasetSQLProperty);

end;

//====================  TUIBDatabaseEditor ==========================

procedure TUIBDatabaseEditor.ExecuteVerb(Index: Integer);
begin
  with TUIBDatabaseEditForm.Create(Application) do
  try
    Database := TUIBDataBase(Component);
    if ShowModal = mrOk then Modified;
  finally
    Free;
  end;
end;

function TUIBDatabaseEditor.GetVerb(Index: Integer): string;
begin
  Result := 'Database Editor ...';
end;

function TUIBDatabaseEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;

//==================== TUIBTransactionEditor ==============================

procedure TUIBTransactionEditor.ExecuteVerb(Index: Integer);
begin
  with TUIBTransactionEditForm.Create(Application) do
  try
    Transaction := TUIBTransaction(Component);
    if ShowModal = mrOk then Modified;
  finally
    Free;
  end;
end;

function TUIBTransactionEditor.GetVerb(Index: Integer): string;
begin
  Result := 'Transaction Editor ...';
end;

function TUIBTransactionEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;

//==================== TUIBSQLEditor ==========================

procedure TUIBSQLEditor.ExecuteVerb(Index: Integer);
begin
  with TUIBSQLEditForm.Create(Application) do
  try
    Statement := TUIBStatement(Component);
    if ShowModal = mrOk then Modified;
  finally
    Free;
  end;
end;

function TUIBSQLEditor.GetVerb(Index: Integer): string;
begin
  Result := 'SQL Editor ...';
end;

function TUIBSQLEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;

//==================== TUIBStatementSQLProperty ========================

procedure TUIBStatementSQLProperty.Edit;
begin
  with TUIBSQLEditForm.Create(Application) do
  try
    Statement := TUIBStatement(GetComponent(0));
    if ShowModal = mrOk then Modified;
  finally
    Free;
  end;
end;

function TUIBStatementSQLProperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paDialog] - [paSubProperties];
end;

//==================== TUIBDatasetSQLProperty ============================

procedure TUIBDatasetSQLProperty.Edit;
var Stm: TUIBStatement;
begin
  with TUIBSQLEditForm.Create(Application) do
  try
    Stm := TUIBStatement.Create(nil);
    try
      Stm.Transaction := TUIBDataSet(GetComponent(0)).Transaction;
      Stm.DataBase := TUIBDataSet(GetComponent(0)).Database;
      Stm.SQL.Assign(TUIBDataSet(GetComponent(0)).SQL);
      Statement := Stm;
      if ShowModal = mrOk then
      begin
        TUIBDataSet(GetComponent(0)).SQL.Assign(Stm.SQL);
        inherited Modified;
      end;
    finally
      Stm.Free;
    end;
  finally
    Free;
  end;
end;

function TUIBDatasetSQLProperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paDialog] - [paSubProperties];
end;


end.

